package zero1hd.rhythmbullet.audio;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.nio.channels.Channels;
import java.security.InvalidParameterException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jaudiotagger.audio.exceptions.InvalidAudioFrameException;
import org.jaudiotagger.audio.exceptions.ReadOnlyFileException;
import org.jaudiotagger.audio.mp3.MP3File;
import org.jaudiotagger.tag.TagException;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.audio.Music.OnCompletionListener;
import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.utils.GdxRuntimeException;

import javazoom.jl.decoder.Bitstream;
import javazoom.jl.decoder.BitstreamException;
import javazoom.jl.decoder.DecoderException;
import javazoom.jl.decoder.Header;
import javazoom.jl.decoder.MP3Decoder;
import javazoom.jl.decoder.OutputBuffer;

public class Mp3Manager implements MusicManager {
	private int readWindowSize = 1024;
	
	private int readWindowIndex;
	private Music playbackMusic;
	private int playbackIndex;
	
	private int sampleRate;
	private long sampleCount;
	private double durationInSeconds;
	private byte channels;
	Bitstream bitstream;
	MP3Decoder decoder;
	OutputBuffer sampleBuffer;
	InputStream inputStream;
	private byte[] currentByteSet;
	private byte[] workset;
	private int indexHead = -1;
	private RandomAccessFile raf;
	private FileHandle file;
	ReentrantLock lock = new ReentrantLock();
	
	private ExecutorService exec;
	
	private String basicSongName;
	public Mp3Manager(FileHandle audioFile) {
		this.file = audioFile;
		lock = new ReentrantLock();
		this.basicSongName = audioFile.name();
		exec = Executors.newSingleThreadExecutor();
		exec.submit(() -> {
			lock.lock();
			Logger.getLogger("org.jaudiotagger").setLevel(Level.OFF);
			try {
				MP3File mp3File = new MP3File(audioFile.file());
				sampleCount = MathUtils.round(Float.valueOf((float) (mp3File.getAudioHeader().getSampleRateAsNumber()*mp3File.getMP3AudioHeader().getPreciseTrackLength()))); 
				sampleRate = mp3File.getMP3AudioHeader().getSampleRateAsNumber();
				durationInSeconds = mp3File.getMP3AudioHeader().getPreciseTrackLength();

			} catch (IOException | TagException | ReadOnlyFileException | InvalidAudioFrameException e) {
				e.printStackTrace();
			}
			lock.unlock();
		});
		
		try {
			raf = new RandomAccessFile(audioFile.file(), "r");
			bitstream = new Bitstream(inputStream = Channels.newInputStream(raf.getChannel()));
			decoder = new MP3Decoder();
		} catch (FileNotFoundException e2) {
			e2.printStackTrace();
			throw new InvalidParameterException("Error with creating RandomAccessFile.");
		}
		try {
			Header header = bitstream.readFrame();
			if (header == null) throw new GdxRuntimeException("Empty MP3");
			channels = (byte) (header.mode() == Header.SINGLE_CHANNEL ? 1 : 2);
			sampleBuffer = new OutputBuffer(channels, false);
			decoder.setOutputBuffer(sampleBuffer);
			workset = new byte[channels*2];
		} catch (BitstreamException e) {
			throw new GdxRuntimeException("error while preloading mp3", e);
		}
		
		playbackMusic = Gdx.audio.newMusic(audioFile);
		
	}
	@Override
	public void setReadIndexToPlaybackIndex() {
		try {
			long pointer = (long) (getPositionInSeconds()*(long)sampleRate)-readWindowSize;
			raf.seek(pointer);
			readWindowIndex = (int) (pointer/readWindowSize);
			currentByteSet = null;
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public void playbackIndexUpdate() {
		playbackIndex = (int) (playbackMusic.getPosition() * sampleRate / readWindowSize);
	}

	@Override
	public int getPlaybackIndexPosition() {
		return playbackIndex;
	}

	@Override
	public int getReadWindowSize() {
		return readWindowSize;
	}

	@Override
	public long getSampleCount() {
		return sampleCount;
	}

	@Override
	public float getDuration() {
		return Float.valueOf(String.valueOf(durationInSeconds));
	}

	@Override
	public int readSamples(float[] samples) {
		int framesRead = 0;
		for (int sid = 0; sid < samples.length; sid++) {
			for (int wsid = 0; wsid < workset.length; wsid++) {
				workset[wsid] = nextByte();
			}
			if (currentByteSet != null) {
				samples[sid] += (workset[1] << 8) + (workset[0] & 0x00ff);
				if (channels > 1) {
					short altChan = (short) ((workset[3] << 8) + (workset[2] & 0x00ff));
					
					samples[sid] = altChan > samples[sid] ? altChan : samples[sid];
				}
				framesRead ++;
				samples[sid] /= Short.MAX_VALUE+1;
			}
		}
		readWindowIndex++;

		return framesRead;
	}
	
	public byte nextByte() {
		indexHead++;
		if (currentByteSet == null || indexHead >= currentByteSet.length) {
			loadNextBuffer();
			if (currentByteSet == null) {
				return 0;
			}
			indexHead = 0;
		}
		
		return currentByteSet[indexHead];
	}
	
	public int loadNextBuffer() {
		if (bitstream != null) {
			int bytesRead = 0;
			try {
				Header header = bitstream.readFrame();
				if (header != null) {
					
					try {
						decoder.decodeFrame(header, bitstream);
					} catch (ArrayIndexOutOfBoundsException | DecoderException e) {
						System.out.println(e);
					}
					
					bitstream.closeFrame();
					bytesRead = sampleBuffer.reset();
					
					currentByteSet = sampleBuffer.getBuffer();
				} else {
					currentByteSet = null;
				}
			} catch (BitstreamException e1) {
				e1.printStackTrace();
			}
			return bytesRead;
		} else {
			return 0;
		}
	}
	
	@Override
	public float getSampleRate() {
		return sampleRate;
	}

	@Override
	public int getreadWindowIndex() {
		return readWindowIndex;
	}

	@Override
	public void pause() {
		Gdx.app.debug("MP3Manager", "Pausing...");
		playbackMusic.pause();
	}

	@Override
	public void play() {
		Gdx.app.debug("Mp3Manager", "MP3 Playing...");
		playbackMusic.play();
	}

	@Override
	public boolean isPlaying() {
		return playbackMusic.isPlaying();
	}

	@Override
	public float getPositionInSeconds() {
		return playbackMusic.getPosition();
	}

	@Override
	public void setPosition(float position) {
		playbackMusic.setPosition(position);
	}

	@Override
	public void setOnCompletionListener(OnCompletionListener listener) {
		playbackMusic.setOnCompletionListener(listener);
	}

	@Override
	public void setVolume(float percent) {
		playbackMusic.setVolume(percent);
	}

	@Override
	public boolean isFinishedLoading() {
		if (lock.isHeldByCurrentThread()) {
			return true;
		} else {
			try {
				if (lock.tryLock(0, TimeUnit.SECONDS)) {
					return true;
				} else {
					return false;
				}
			} catch (InterruptedException e) {
				return false;
			}
		}
	}

	@Override
	public String getBasicSongName() {
		return basicSongName;
	}

	@Override
	public FileHandle getMusicFile() {
		return file;
	}
	
	@Override
	public void dispose() {
		Gdx.app.debug("MP3Manager", "Disposing...");
		playbackMusic.stop();
		playbackMusic.dispose();
		exec.shutdown();
		try {
			bitstream.close();
			bitstream = null;
			inputStream.close();
		} catch (BitstreamException | IOException e) {
			e.printStackTrace();
		}
	}
}
